@@ -45,7 +45,6 @@
 
 #include "libguile/async.h"
 #include "libguile/eval.h"
-#include "libguile/root.h"
 #include "libguile/vectors.h"
 #include "libguile/threads.h"
 
@@ -86,6 +85,8 @@
    signal_handler_threads points to the thread that a signal should be
    delivered to.
 */
+static scm_i_pthread_mutex_t signal_handler_lock =
+  SCM_I_PTHREAD_MUTEX_INITIALIZER;
 static SCM *signal_handlers;
 static SCM signal_handler_asyncs;
 static SCM signal_handler_threads;
@@ -109,8 +110,10 @@ static SIGRETTYPE (*orig_handlers[NSIG])(int);
 static SCM
 close_1 (SCM proc, SCM arg)
 {
-  return scm_primitive_eval_x (scm_list_3 (scm_sym_lambda, SCM_EOL,
-					   scm_list_2 (proc, arg)));
+  /* Eval in the root module so that `lambda' has its usual meaning.  */
+  return scm_eval (scm_list_3 (scm_sym_lambda, SCM_EOL,
+                               scm_list_2 (proc, arg)),
+                   scm_the_root_module ());
 }
 
 #if SCM_USE_PTHREAD_THREADS
@@ -159,6 +162,11 @@ signal_delivery_thread (void *data)
 #if HAVE_PTHREAD_SIGMASK  /* not on mingw, see notes above */
   sigset_t all_sigs;
   sigfillset (&all_sigs);
+  /* On libgc 7.1 and earlier, GC_do_blocking doesn't actually do
+     anything.  So in that case, libgc will want to suspend the signal
+     delivery thread, so we need to allow it to do so by unmasking the
+     suspend signal.  */
+  sigdelset (&all_sigs, GC_get_suspend_signal ());
   scm_i_pthread_sigmask (SIG_SETMASK, &all_sigs, NULL);
 #endif
 
@@ -221,9 +229,8 @@ take_signal (int signum)
 
   if (scm_is_false (SCM_CDR (cell)))
     {
-      SCM_SETCDR (cell, t->active_asyncs);
-      t->active_asyncs = cell;
-      t->pending_asyncs = 1;
+      SCM_SETCDR (cell, t->pending_asyncs);
+      t->pending_asyncs = cell;
     }
 
 #ifndef HAVE_SIGACTION
@@ -320,15 +327,14 @@ SCM_DEFINE (scm_sigaction_for_thread, "sigaction", 1, 3, 0,
   if (SCM_UNBNDP (thread))
     thread = scm_current_thread ();
   else
-    {
-      SCM_VALIDATE_THREAD (4, thread);
-      if (scm_c_thread_exited_p (thread))
-	SCM_MISC_ERROR ("thread has already exited", SCM_EOL);
-    }
+    SCM_VALIDATE_THREAD (4, thread);
 
   scm_i_ensure_signal_delivery_thread ();
 
-  SCM_CRITICAL_SECTION_START;
+  scm_dynwind_begin (0);
+  scm_i_dynwind_pthread_mutex_lock (&signal_handler_lock);
+  scm_dynwind_block_asyncs ();
+
   old_handler = SCM_SIMPLE_VECTOR_REF (*signal_handlers, csig);
   if (SCM_UNBNDP (handler))
     query_only = 1;
@@ -347,7 +353,6 @@ SCM_DEFINE (scm_sigaction_for_thread, "sigaction", 1, 3, 0,
 	}
       else
 	{
-	  SCM_CRITICAL_SECTION_END;
 	  SCM_OUT_OF_RANGE (2, handler);
 	}
     }
@@ -434,7 +439,9 @@ SCM_DEFINE (scm_sigaction_for_thread, "sigaction", 1, 3, 0,
     }
   if (old_action.sa_handler == SIG_DFL || old_action.sa_handler == SIG_IGN)
     old_handler = scm_from_long ((long) old_action.sa_handler);
-  SCM_CRITICAL_SECTION_END;
+
+  scm_dynwind_end ();
+
   return scm_cons (old_handler, scm_from_int (old_action.sa_flags));
 #else
   if (query_only)
@@ -453,7 +460,9 @@ SCM_DEFINE (scm_sigaction_for_thread, "sigaction", 1, 3, 0,
     }
   if (old_chandler == SIG_DFL || old_chandler == SIG_IGN)
     old_handler = scm_from_long ((long) old_chandler);
-  SCM_CRITICAL_SECTION_END;
+
+  scm_dynwind_end ();
+
   return scm_cons (old_handler, scm_from_int (0));
 #endif
 }
@@ -508,6 +517,23 @@ SCM_DEFINE (scm_alarm, "alarm", 1, 0, 0,
 #undef FUNC_NAME
 #endif /* HAVE_ALARM */
 
+static void
+pack_tv (struct timeval *tv, SCM seconds, SCM microseconds)
+{
+  tv->tv_sec = scm_to_long (seconds);
+  tv->tv_usec = scm_to_long (microseconds);
+
+  /* Allow usec to be outside the range [0, 999999).  */
+  tv->tv_sec += tv->tv_usec / (1000 * 1000);
+  tv->tv_usec %= 1000 * 1000;
+}
+
+static SCM
+unpack_tv (const struct timeval *tv)
+{
+  return scm_cons (scm_from_long (tv->tv_sec), scm_from_long (tv->tv_usec));
+}
+
 #ifdef HAVE_SETITIMER
 SCM_DEFINE (scm_setitimer, "setitimer", 5, 0, 0,
            (SCM which_timer,
@@ -537,20 +563,16 @@ SCM_DEFINE (scm_setitimer, "setitimer", 5, 0, 0,
   struct itimerval old_timer;
 
   c_which_timer = SCM_NUM2INT(1, which_timer);
-  new_timer.it_interval.tv_sec = SCM_NUM2LONG(2, interval_seconds);
-  new_timer.it_interval.tv_usec = SCM_NUM2LONG(3, interval_microseconds);
-  new_timer.it_value.tv_sec = SCM_NUM2LONG(4, value_seconds);
-  new_timer.it_value.tv_usec = SCM_NUM2LONG(5, value_microseconds);
+  pack_tv (&new_timer.it_interval, interval_seconds, interval_microseconds);
+  pack_tv (&new_timer.it_value, value_seconds, value_microseconds);
 
   SCM_SYSCALL(rv = setitimer(c_which_timer, &new_timer, &old_timer));
   
   if(rv != 0)
     SCM_SYSERROR;
 
-  return scm_list_2 (scm_cons (scm_from_long (old_timer.it_interval.tv_sec),
-			       scm_from_long (old_timer.it_interval.tv_usec)),
-		     scm_cons (scm_from_long (old_timer.it_value.tv_sec),
-			       scm_from_long (old_timer.it_value.tv_usec)));
+  return scm_list_2 (unpack_tv (&old_timer.it_interval),
+                     unpack_tv (&old_timer.it_value));
 }
 #undef FUNC_NAME
 #endif /* HAVE_SETITIMER */
